/***************************************************************************
    cZ11.c  -  driver for the Lexmark Z11 printer. See README for details.
                             -------------------
    begin                : Mon Jun 12 14:07:33 CEST 2000
    last change          : Sun Jun 18 16:05:00 CEST 2000
    copyright            : (C) 2000 by Christian Kornblum
    email                : kornblum@studst.fh-muenster.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>

#include <stdlib.h>
#include <string.h>

/* various constants */
#define Z11_PROLOG "/etc/LexmarkZ11/prolog.prn"
#define MAX_LINES 600
#define PAGE_HEIGHT 6720
#define PAGE_WIDTH 4802
#define GS_PAGE_WIDTH 4958
#define GS_PAGE_HEIGHT 7017

/* general */
#define ROW_GAP 16
#define BYTES_PER_HEADER 24
#define LEFT_MARGIN 10
#define UPPER_MARGIN 0

/* CMYK printing */
#define COLOR_PENS 64
#define COLOR_GAP 26 /* ? */
#define BYTES_PER_COLUMN 24
#define STD_PAGE_MOVE 2*COLOR_PENS
#define COLOR_BUFFERS 6

/* the ghostscript color identifiers */
#define BLACK   0x10
#define CYAN    0x80
#define MAGENTA 0x40
#define YELLOW  0x20

/* the structure for the pixmaps */
struct tSweepBuffer {
  int bytepos;
  int bitpos;
  int bufpos;        /* this is used for the colors! */
  int unprinted;     /* does this buffer contain data? */
  char *buffer;
};

/*
 * This writes a number of zeros to a string.
 */
void ClearBuffer(char *data, int bytes)
{
    memset((void *) data, 0, bytes);
} /* ClearBuffer */

/*
 * Initialize a sweep buffer
 */
void SweepBuffer_Init (struct tSweepBuffer *SweepBuffer, int bytesize)
{
  SweepBuffer->bytepos = 0;
  SweepBuffer->bitpos = 0;
  SweepBuffer->bufpos = 0;
  SweepBuffer->unprinted = 0;
  SweepBuffer->buffer = (char *) malloc(bytesize+10000);
  ClearBuffer(SweepBuffer->buffer, bytesize+10000);
} /* SweepBuffer_Init */

/*
 * Summ up some bytes for a single checksum-byte
 */
char checkbyte(char* stuff, int start, int end)
{
    int i;
    char ret = 0;
    for(i = start; i <= end; i++) ret += stuff[i];
    return ret;
} /* CheckByte */

/*
 * This puts an unterminated amount of any chars to "out". The first
 * byte of the "string" has to give the correct number of the following bytes.
 */
void fPutLString (FILE *out, char *data) {
  int i;
  for (i = 1; i <= data[0]; putc(data[i++], out));
} /* fPutLString */

/*
 * This moves the paper by a defined number of lines (600lpi!).
 */
void LexMove(FILE *out, long int pixel)
{
  char command[] = {8, 0x1b,0x23,0x80,0x0a,0x58,0x00,0x00,0x05,0x1b,0x31,0x0c,0x02,0x7f,0x00,0x00,0xbe};
  pixel *= 2;
  command[5] = (char) pixel;
  command[4] = (char) (pixel >> 8);
  command[8] = checkbyte(command, 2, 7);
  fPutLString(out, command);
} /* LexMove */

/*
 * This initializes the printer. Currently, the rather long magic
 * prolog is read from a file.
 */
void LexInit(FILE *out)
{
    int InByte;
    FILE* prologfile;

    /* The file containig the prolog should be there... */
    if (!(prologfile = fopen(Z11_PROLOG, "r")))
    {
        fprintf(stderr, "LexZ11: Cannot find prolog file '%s'. Aborting.\n\n", Z11_PROLOG);
        exit(-1);
    }
    while ((InByte = getc(prologfile)) != EOF)
        putc(InByte, out);

    fclose(prologfile);
} /* LexInit */

/*
 * This tells the printer to throw out his current page.
 */
void LexEOP(FILE *out)
{
   char command[] = {24, 0x1B,0x22,0x80,0x3f,0x84,0x00,0x00,0x65,0x1b,0x33,0x00,0x00,0x00,0x00,0x00,
                        0x33,0x1b,0x62,0x00,0x00,0x00,0x00,0x00,0x62};
   fPutLString(out, command);
}

/*
 * Make bytes to use both inkjet columns on the cartridge. Return 1 if
 * they are not empty.
 */
int MakePacketByte(char *buffer, int bytesize, int index, char *packet, int bytespercolumn)
{
  int rowindex;
  *packet = 0;

  // calculate position of data byte for second row
  rowindex = index - (bytespercolumn * ROW_GAP);

  // set up byte
  if (index < bytesize)
    *packet = buffer[index] & 0x55;
  if ((rowindex >= 0) && (rowindex < bytesize))
    *packet |= buffer[rowindex] & 0xAA;

  return *packet;

} /* MakePacketBytes */

/*
 * Make packets to reduce the bytes sent to the printer. This is done
 * with a 24 bit mask and up to 24 bytes for the pixels.
 * The 16 pixel space between the two columns on the cartridge has
 * to be taken into account, too.
 */
void MakePackets(char *buffer, int colstart, int *width,
		 char *packetbuffer, int *bytesize, int bytespercolumn)
{
  int packbufpos = 0;
  int lastpackheadpos = 0;
  char mergebit, packet;
  char *columnbuffer;
  register int i, j;

  /* less calculating */
  columnbuffer = &buffer[colstart];

  /* do things reverse */
  for (j= *width + ROW_GAP - 1; j >= 0; j--)
  {
    lastpackheadpos = packbufpos;
    packbufpos += 4;
    packetbuffer[lastpackheadpos] = 0x80;
    packetbuffer[lastpackheadpos + 1] = 0xff;
	packetbuffer[lastpackheadpos + 2] = 0xff;
	packetbuffer[lastpackheadpos + 3] = 0xff;
	
    for (i = 0; i < bytespercolumn; i++)
    {
       /* The gap is crap. Just to save a uP... a shame. */
        if (MakePacketByte(columnbuffer, *bytesize, i + j*bytespercolumn, &packet, bytespercolumn))
	    {
	        /* shift the bit to the packet position and activate it */
            mergebit = 0x80 >> (i % 8);
            packetbuffer[lastpackheadpos + i / 8 + 1] ^= mergebit;

            /* add byte to the buffer */	
	        packetbuffer[packbufpos] = packet;
	        packbufpos++;
	    }
	}

  }

  /* new bytesize */
  *bytesize = packbufpos;

} /* MakePackets */


/*
 * This sends a complete sweep to the printer. Black or color, no difference.
 */
void PrintSweep(char *buffer, char *header, int bytesize,
		int width, FILE *out, int bytespercolumn)
{
  int bstart = 0;
  int leftmargin = LEFT_MARGIN;
  char *packetbuffer;
  int packetbuffersize;
  register int i;

  /* Allocate Packetbuffer */
  packetbuffersize = (PAGE_WIDTH + 2 * ROW_GAP) * (bytespercolumn + 4) + 20000;
  packetbuffer = (char *) malloc(packetbuffersize);
  memset(packetbuffer, 0, packetbuffersize);

  if (width > 0) { /* do not print empty sweeps */
    /* kill the noise at the end of the lines... probably MY bug */
    width += ROW_GAP;
    bytesize += BYTES_PER_COLUMN * ROW_GAP;

    /* Make the packets for the printer. */
    MakePackets(buffer, bstart, &width, packetbuffer, &bytesize, bytespercolumn);

    /* Calculate the number of bytes for the size of the pixelblock */
    header[20] = (char) bytesize;
    header[19] = (char) (bytesize >> 8);
    header[18] = (char) (bytesize >> 16);

    /* The number of columns */
    header[4] = (char) width;
    header[3] = (char) (width >> 8);

    /* vertical move, always just enough for the next color */
    header[14] = COLOR_PENS * 2;
    header[13] = 0;

    /* checksum-bytes - this did cost me some nerve! */
    header[7] = checkbyte(header, 1, 6);
    header[15] = checkbyte(header, 9, 14);
    header[23] = checkbyte(header, 17, 22);

    /* send header and compressed matrix to the printer */
    fwrite((void *) header, 1, BYTES_PER_HEADER, out);
    fwrite((void *) packetbuffer, 1, bytesize, out);
  }

  free(packetbuffer);

} /* PrintSweep */	

/*
 * This finds out if there is anything but zeros in a string
 */
int LineSum(char line[], int length)
{
  register int i = 0;
  while (i < length)
    if(line[i++] != 0) return 1;
  return 0;
} /* LineSum */

/*
 * This is the main printing routine. Wicked and insane. Nonetheless working.
 * This one is printing CMYK and expects the ghostscript "bitcmyk" device.
 */
void LexPrintCMYK(FILE *in, FILE *out) {
  char line[GS_PAGE_WIDTH / 2];
  int done_page, cur_height = 0, page_height = 0, numpages = 0;
  char lex_colhd[BYTES_PER_HEADER] = {  0x1b,0x40,0x80,0x12,0xd2,0x04,0xb7,0xFF,
				                        0x1b,0x42,0x0c,0x04,0x8f,0x00,0xFF,0xFF,
				                        0x1b,0x43,0x00,0xFF,0xFF,0x39,0x87,0xFF };

  long bytesize;
  register int i=0;
  int CurrentColBuffer = 0, BufferToPrint;
  int width;
  int empty_lines;
  char nibble;
  struct tSweepBuffer colbuffer[COLOR_BUFFERS];

  /* The printer may not be able to print every GhostScript pixel */
  if (GS_PAGE_WIDTH <= PAGE_WIDTH) width = GS_PAGE_WIDTH;
  else width = PAGE_WIDTH;

  /* Calculating the size for the buffers */
  bytesize = BYTES_PER_COLUMN * width;

  /* As long as we get input... */
  while((line[0] = getc(in)) != EOF)
  {

    /* Get memory and clear it. */
    for (i=0; i<COLOR_BUFFERS; i++) {
      SweepBuffer_Init(&colbuffer[i], bytesize);
      colbuffer[i].bufpos = i;
    }

    /* Initialize the printer, load a page  */
    LexInit(out);
    LexMove(out, UPPER_MARGIN);

    /* Reset all variables */
    done_page   = 0;
    page_height = 0;
    cur_height  = 0;
    empty_lines = 0;

    /* ... we do the pages. */
    while(!done_page)
    {

      /* Read a CMYK line (GS -sDEVICE=bitcmyk) from the input */
      if (page_height == 0) {
        fread((void *) &line[1], 1, GS_PAGE_WIDTH / 2 - 1, in);
      }
      else {
	    fread((void *) &line[0], 1, GS_PAGE_WIDTH / 2, in);
      }

      /* optimize for empty lines, if buffers are empty */
      if ((cur_height == 0)
	  && !LineSum(line, GS_PAGE_WIDTH / 2)
	  && (page_height < PAGE_HEIGHT)
	  && (page_height < GS_PAGE_HEIGHT)
	  && !(colbuffer[0].unprinted | colbuffer[1].unprinted | colbuffer[3].unprinted))
	{
	  empty_lines++;
	}
      else /* This line does not seem to be empty or there is still data */
	{
	  if (empty_lines) {
        LexMove(out, empty_lines);
        empty_lines = 0;	
	  }

	  /* count lines and set values */
	
	  /* yellow */
	  colbuffer[0].bitpos  = 7 - (cur_height % 8);
	  colbuffer[0].bytepos = 16 + (cur_height / 8) % 8;
	  colbuffer[0].bufpos  = cur_height / COLOR_PENS;
	
	  /* magenta */
	  colbuffer[1].bitpos
	    = 7 - ((cur_height + COLOR_GAP + COLOR_PENS) % 8);
	  colbuffer[1].bytepos
	    = 8 + ((cur_height + COLOR_GAP + COLOR_PENS) / 8) % 8;
	  colbuffer[1].bufpos
	    = ((cur_height + COLOR_GAP + COLOR_PENS) / COLOR_PENS) % 3;

	  /* cyan has 6 buffers, so that it is not mapped to buffers
	     which have not been printed by yellow yet. The Buffers
	     > 2 are mapped to the right corresponding buffer
	     after it has been sent to the printer. */
	  colbuffer[2].bitpos
	    = 7 - ((cur_height + 2 * (COLOR_GAP + COLOR_PENS)) % 8);
	  colbuffer[2].bytepos
	    = ((cur_height + 2 * (COLOR_GAP + COLOR_PENS)) / 8) % 8;
	  colbuffer[2].bufpos
	    = ((cur_height + 2 * (COLOR_GAP + COLOR_PENS)) / COLOR_PENS) % 3;
	  if (colbuffer[2].bufpos == colbuffer[0].bufpos)
	    colbuffer[2].bufpos += 3;

	  /* This extracts the nibbles and transforms them to the bits
	     in the output stream. */
	  for(i=0; (i <= width); i++)
	    {
	      nibble = (line[i/2] << (4 * (i % 2))) & 0xF0;
	      if (nibble & BLACK) {	
            nibble = CYAN | MAGENTA | YELLOW; /* this is supposed to be black, then. */	
	      }
	      if (nibble & CYAN) {
		    colbuffer[colbuffer[2].bufpos].buffer[(i * BYTES_PER_COLUMN) + colbuffer[2].bytepos]
        	  |= 0x01 << colbuffer[2].bitpos;
		    colbuffer[colbuffer[2].bufpos].unprinted = 1;
	      }
	      if (nibble & MAGENTA) {
		    colbuffer[colbuffer[1].bufpos].buffer[(i * BYTES_PER_COLUMN) + colbuffer[1].bytepos]
    		  |= 0x01 << colbuffer[1].bitpos;
	    	colbuffer[colbuffer[1].bufpos].unprinted = 1;
	      }
	      if (nibble & YELLOW) {
		    colbuffer[colbuffer[0].bufpos].buffer[(i * BYTES_PER_COLUMN) + colbuffer[0].bytepos]
    		  |= 0x01 << colbuffer[0].bitpos;
		    colbuffer[colbuffer[0].bufpos].unprinted = 1;
	      }
	    }
	  cur_height++;
	  /* Buffer is full or page is over. Print it. Color first...*/
	  if (!(cur_height % COLOR_PENS) || (page_height >= GS_PAGE_HEIGHT))
	    {
	      PrintSweep(colbuffer[CurrentColBuffer].buffer,
			 lex_colhd, bytesize, width, out, BYTES_PER_COLUMN);
	      ClearBuffer(colbuffer[CurrentColBuffer].buffer, bytesize);
 	      colbuffer[CurrentColBuffer].unprinted = 0;
 	
	      /* now handle the cyan stuff */
	      if (colbuffer[CurrentColBuffer + 3].unprinted)
	      {
	        for(i = 0; i < bytesize; i++)
		        colbuffer[CurrentColBuffer].buffer[i] |= colbuffer[CurrentColBuffer + 3].buffer[i];
	        ClearBuffer(colbuffer[CurrentColBuffer + 3].buffer, bytesize);
	        colbuffer[CurrentColBuffer].unprinted = colbuffer[CurrentColBuffer + 3].unprinted;
	
	        colbuffer[CurrentColBuffer + 3].unprinted = 0;
	      }
	
	      /* switch to the next buffer */
	      CurrentColBuffer = ++CurrentColBuffer % 3;
	
	    }
  	  if (cur_height == 3 * COLOR_PENS) cur_height = 0;
	}

      /* this page has found an end */
      if ((page_height++ >= PAGE_HEIGHT)||(page_height >= GS_PAGE_HEIGHT))
      {
        done_page = 1;
        /* print all unprinted buffers */
        for(i = 0; i < COLOR_BUFFERS; i++)
        {
            BufferToPrint = (i + CurrentColBuffer) % 3;

            if(colbuffer[BufferToPrint].unprinted)
            {
                fprintf(stderr, "printing post-sweep %d\n", BufferToPrint);
	            PrintSweep(colbuffer[BufferToPrint].buffer,
			        lex_colhd, bytesize, width, out, BYTES_PER_COLUMN);
			    colbuffer[BufferToPrint].unprinted = 0;
	        }
	
	        /* now handle the cyan stuff */
	        if (colbuffer[BufferToPrint + 3].unprinted)
	        {
	            for(i = 0; i < bytesize; i++)
		            colbuffer[BufferToPrint].buffer[i] |= colbuffer[BufferToPrint + 3].buffer[i];
	            ClearBuffer(colbuffer[BufferToPrint + 3].buffer, bytesize);
	
	            colbuffer[BufferToPrint].unprinted = colbuffer[BufferToPrint + 3].unprinted;
	
	            colbuffer[BufferToPrint + 3].unprinted = 0;
	        }
        }
      }

    }

    /* hand out the page */
    LexEOP(out);

    /* eat any remaining whitespace so process will not hang */
    if (PAGE_HEIGHT < GS_PAGE_HEIGHT)
      for(i=0; i < ((GS_PAGE_HEIGHT - PAGE_HEIGHT) * GS_PAGE_WIDTH / 2); i++) getc(in);

    /* count the pages and free memory */
    numpages++;
    for (i=0; i < COLOR_BUFFERS; free(colbuffer[i++].buffer));
  }
  if (numpages == 0) fprintf(stderr, "LexZ11: No pages printed!");
} /* LexPrintCMYK */

#ifdef DEBUG
/*
 * Write input to a file. Only for debugging purposes.
 */
void WriteToFile(FILE *in, FILE *out)
{
  char c;
  while ((c = getc(in)) != EOF) putc(c, out);
} /* WriteToFile */
#endif

/*
 * The main program. Sets input and output streams.
 */
int main(int argc, char *argv[]) {
    FILE *InputFile;
    FILE *OutPutFile;

    InputFile  = stdin;
    OutPutFile = stdout;


#ifdef DEBUG
    printf("debug mode!\n");
    OutPutFile = fopen("/tmp/Z11out", "w");
    DebugFile = fopen("/tmp/Z11debug", "w");
#endif

    LexPrintCMYK(InputFile, OutPutFile);

    fclose(OutPutFile);
    fclose(InputFile);
#ifdef DEBUG
    fclose(DebugFile);
#endif
    return 0;
}
